# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

A2UI_ASSET_PACKAGE = "a2ui.assets"
SERVER_TO_CLIENT_SCHEMA_KEY = "server_to_client"
COMMON_TYPES_SCHEMA_KEY = "common_types"
CATALOG_SCHEMA_KEY = "catalog"
CATALOG_COMPONENTS_KEY = "components"
CATALOG_ID_KEY = "catalogId"
CATALOG_STYLES_KEY = "styles"
SURFACE_ID_KEY = "surfaceId"

# Protocol constants
SUPPORTED_CATALOG_IDS_KEY = "supportedCatalogIds"
INLINE_CATALOGS_KEY = "inlineCatalogs"

A2UI_CLIENT_CAPABILITIES_KEY = "a2uiClientCapabilities"

BASE_SCHEMA_URL = "https://a2ui.org/"
INLINE_CATALOG_NAME = "inline"

VERSION_0_8 = "0.8"
VERSION_0_9 = "0.9"
VERSION_0_9_1 = "0.9.1"

SPEC_VERSION_MAP = {
    VERSION_0_8: {
        SERVER_TO_CLIENT_SCHEMA_KEY: "specification/v0_8/json/server_to_client.json",
    },
    VERSION_0_9: {
        SERVER_TO_CLIENT_SCHEMA_KEY: "specification/v0_9/json/server_to_client.json",
        COMMON_TYPES_SCHEMA_KEY: "specification/v0_9/json/common_types.json",
    },
    VERSION_0_9_1: {
        SERVER_TO_CLIENT_SCHEMA_KEY: "specification/v0_9_1/json/server_to_client.json",
        COMMON_TYPES_SCHEMA_KEY: "specification/v0_9_1/json/common_types.json",
    },
}

SPECIFICATION_DIR = "specification"

ENCODING = "utf-8"

A2UI_OPEN_TAG = "<a2ui-json>"
A2UI_CLOSE_TAG = "</a2ui-json>"

A2UI_SCHEMA_BLOCK_START = "---BEGIN A2UI JSON SCHEMA---"
A2UI_SCHEMA_BLOCK_END = "---END A2UI JSON SCHEMA---"

DEFAULT_WORKFLOW_RULES = f"""
The generated response MUST follow these rules:
- The response can contain one or more A2UI JSON blocks.
- Each A2UI JSON block MUST be wrapped in `{A2UI_OPEN_TAG}` and `{A2UI_CLOSE_TAG}` tags.
- Between or around these blocks, you can provide conversational text.
- The JSON part MUST be a single, raw JSON object (usually a list of A2UI messages) and MUST validate against the provided A2UI JSON SCHEMA.
- Top-Down Component Ordering: Within the `components` list of a message:
    - The 'root' component MUST be the FIRST element.
    - Parent components MUST appear before their child components.
    This specific ordering allows the streaming parser to yield and render the UI incrementally as it arrives.
"""


# A2UI Tool constants
A2UI_TOOL_NAME = "send_a2ui_json_to_client"
A2UI_VALIDATED_JSON_KEY = "validated_a2ui_json"
A2UI_TOOL_ERROR_KEY = "error"



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest.mock import MagicMock

import pytest

from a2ui.adk.send_a2ui_to_client_toolset import SendA2uiToClientToolset
from a2ui.schema.catalog import A2uiCatalog
from google.adk.agents.readonly_context import ReadonlyContext
from google.adk.tools.tool_context import ToolContext
from google.genai import types as genai_types

# region SendA2uiToClientToolset Tests
"""Tests for the SendA2uiToClientToolset class."""


@pytest.mark.asyncio
async def test_toolset_init_bool():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  toolset = SendA2uiToClientToolset(
      a2ui_enabled=True, a2ui_catalog=catalog_mock, a2ui_examples="examples"
  )
  ctx = MagicMock(spec=ReadonlyContext)
  assert await toolset._resolve_a2ui_enabled(ctx) == True

  # Access the tool to check schema resolution
  tool = toolset._ui_tools[0]
  assert await tool._resolve_a2ui_catalog(ctx) == catalog_mock


@pytest.mark.asyncio
async def test_toolset_init_callable():
  enabled_mock = MagicMock(return_value=True)
  catalog_mock = MagicMock(spec=A2uiCatalog)
  examples_mock = MagicMock(return_value="examples")
  toolset = SendA2uiToClientToolset(
      a2ui_enabled=enabled_mock,
      a2ui_catalog=catalog_mock,
      a2ui_examples=examples_mock,
  )
  ctx = MagicMock(spec=ReadonlyContext)
  assert await toolset._resolve_a2ui_enabled(ctx) == True

  # Access the tool to check schema resolution
  tool = toolset._ui_tools[0]
  assert await tool._resolve_a2ui_catalog(ctx) == catalog_mock
  assert await tool._resolve_a2ui_examples(ctx) == "examples"
  enabled_mock.assert_called_once_with(ctx)
  catalog_mock.assert_not_called()  # It's an object, not a callable in this test
  examples_mock.assert_called_once_with(ctx)


@pytest.mark.asyncio
async def test_toolset_init_async_callable():
  async def async_enabled(_ctx):
    return True

  catalog_mock = MagicMock(spec=A2uiCatalog)

  async def async_catalog(_ctx):
    return catalog_mock

  async def async_examples(_ctx):
    return "examples"

  toolset = SendA2uiToClientToolset(
      a2ui_enabled=async_enabled,
      a2ui_catalog=async_catalog,
      a2ui_examples=async_examples,
  )
  ctx = MagicMock(spec=ReadonlyContext)
  assert await toolset._resolve_a2ui_enabled(ctx) == True

  # Access the tool to check schema resolution
  tool = toolset._ui_tools[0]
  assert await tool._resolve_a2ui_catalog(ctx) == catalog_mock
  assert await tool._resolve_a2ui_examples(ctx) == "examples"


@pytest.mark.asyncio
async def test_toolset_get_tools_enabled():
  toolset = SendA2uiToClientToolset(
      a2ui_enabled=True, a2ui_catalog=MagicMock(spec=A2uiCatalog), a2ui_examples=""
  )
  tools = await toolset.get_tools(MagicMock(spec=ReadonlyContext))
  assert len(tools) == 1
  assert isinstance(tools[0], SendA2uiToClientToolset._SendA2uiJsonToClientTool)


@pytest.mark.asyncio
async def test_toolset_get_tools_disabled():
  toolset = SendA2uiToClientToolset(
      a2ui_enabled=False,
      a2ui_catalog=MagicMock(spec=A2uiCatalog),
      a2ui_examples="",
  )
  tools = await toolset.get_tools(MagicMock(spec=ReadonlyContext))
  assert len(tools) == 0


# endregion

# region SendA2uiJsonToClientTool Tests
"""Tests for the _SendA2uiJsonToClientTool class."""


def test_send_tool_init():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")
  assert tool.name == SendA2uiToClientToolset._SendA2uiJsonToClientTool.TOOL_NAME
  assert tool._a2ui_catalog == catalog_mock
  assert tool._a2ui_examples == "examples"


def test_send_tool_get_declaration():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")
  declaration = tool._get_declaration()
  assert declaration is not None
  assert declaration.name == SendA2uiToClientToolset._SendA2uiJsonToClientTool.TOOL_NAME
  assert (
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME
      in declaration.parameters.properties
  )
  assert (
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME
      in declaration.parameters.required
  )


@pytest.mark.asyncio
async def test_send_tool_resolve_catalog():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")
  catalog = await tool._resolve_a2ui_catalog(MagicMock(spec=ReadonlyContext))
  assert catalog == catalog_mock


@pytest.mark.asyncio
async def test_send_tool_resolve_examples():
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(
      MagicMock(spec=A2uiCatalog), "examples"
  )
  examples = await tool._resolve_a2ui_examples(MagicMock(spec=ReadonlyContext))
  assert examples == "examples"


@pytest.mark.asyncio
async def test_send_tool_process_llm_request():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  catalog_mock.render_as_llm_instructions.return_value = "rendered_catalog"
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")

  tool_context_mock = MagicMock(spec=ToolContext)
  tool_context_mock.state = {}
  llm_request_mock = MagicMock()
  llm_request_mock.append_instructions = MagicMock()

  await tool.process_llm_request(
      tool_context=tool_context_mock, llm_request=llm_request_mock
  )

  llm_request_mock.append_instructions.assert_called_once()
  args, _ = llm_request_mock.append_instructions.call_args
  instructions = args[0]
  assert "rendered_catalog" in instructions
  assert "examples" in instructions


@pytest.mark.asyncio
async def test_send_tool_run_async_valid():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")
  tool_context_mock = MagicMock(spec=ToolContext)
  tool_context_mock.state = {}
  tool_context_mock.actions = MagicMock(skip_summarization=False)

  valid_a2ui = [{"type": "Text", "text": "Hello"}]
  catalog_mock.validator.validate.return_value = None
  args = {
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME: json.dumps(
          valid_a2ui
      )
  }

  result = await tool.run_async(args=args, tool_context=tool_context_mock)
  assert result == {
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.VALIDATED_A2UI_JSON_KEY: (
          valid_a2ui
      )
  }
  assert tool_context_mock.actions.skip_summarization == True
  catalog_mock.validator.validate.assert_called_once_with(valid_a2ui)


@pytest.mark.asyncio
async def test_send_tool_run_async_valid_list():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")
  tool_context_mock = MagicMock(spec=ToolContext)
  tool_context_mock.state = {}
  tool_context_mock.actions = MagicMock(skip_summarization=False)

  valid_a2ui = [{"type": "Text", "text": "Hello"}]
  catalog_mock.validator.validate.return_value = None
  args = {
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME: json.dumps(
          valid_a2ui
      )
  }

  result = await tool.run_async(args=args, tool_context=tool_context_mock)
  assert result == {
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.VALIDATED_A2UI_JSON_KEY: (
          valid_a2ui
      )
  }
  assert tool_context_mock.actions.skip_summarization == True
  catalog_mock.validator.validate.assert_called_once_with(valid_a2ui)


@pytest.mark.asyncio
async def test_send_tool_run_async_missing_arg():
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(
      MagicMock(spec=A2uiCatalog), "examples"
  )
  result = await tool.run_async(args={}, tool_context=MagicMock())
  assert "error" in result
  assert (
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME
      in result["error"]
  )


@pytest.mark.asyncio
async def test_send_tool_run_async_invalid_json():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")
  args = {
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME: "{invalid"
  }
  result = await tool.run_async(args=args, tool_context=MagicMock())
  assert "error" in result
  assert "Failed to call A2UI tool" in result["error"]
  assert "Expecting property name enclosed in double quotes" in result["error"]


@pytest.mark.asyncio
async def test_send_tool_run_async_schema_validation_fail():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  catalog_mock.validator.validate.side_effect = Exception(
      "'text' is a required property"
  )
  tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")
  invalid_a2ui = [{"type": "Text"}]  # Missing 'text'
  args = {
      SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME: json.dumps(
          invalid_a2ui
      )
  }
  result = await tool.run_async(args=args, tool_context=MagicMock())
  assert "error" in result
  assert "Failed to call A2UI tool" in result["error"]
  assert "'text' is a required property" in result["error"]


# endregion



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the A2uiPartConverter class."""

import json
from unittest.mock import MagicMock, patch

import pytest

from a2a import types as a2a_types
from a2ui.a2a.parts import create_a2ui_part
from a2ui.adk.a2a.part_converter import A2uiPartConverter
from a2ui.adk.send_a2ui_to_client_toolset import SendA2uiToClientToolset
from a2ui.schema.catalog import A2uiCatalog
from a2ui.schema.constants import A2UI_CLOSE_TAG, A2UI_OPEN_TAG
from google.genai import types as genai_types


def test_converter_class_convert_valid_tool_response():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  valid_a2ui = {"type": "Text", "text": "Hello"}
  function_response = genai_types.FunctionResponse(
      name=SendA2uiToClientToolset._SendA2uiJsonToClientTool.TOOL_NAME,
      response={
          SendA2uiToClientToolset._SendA2uiJsonToClientTool.VALIDATED_A2UI_JSON_KEY: [
              valid_a2ui
          ]
      },
  )
  part = genai_types.Part(function_response=function_response)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 1
  assert a2a_parts[0] == create_a2ui_part(valid_a2ui)


def test_converter_class_convert_tool_error_response():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  function_response = genai_types.FunctionResponse(
      name=SendA2uiToClientToolset._SendA2uiJsonToClientTool.TOOL_NAME,
      response={
          SendA2uiToClientToolset._SendA2uiJsonToClientTool.TOOL_ERROR_KEY: "Some error"
      },
  )
  part = genai_types.Part(function_response=function_response)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 0


def test_converter_class_convert_tool_response_no_result():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  function_response = genai_types.FunctionResponse(
      name=SendA2uiToClientToolset._SendA2uiJsonToClientTool.TOOL_NAME,
      response={},
  )
  part = genai_types.Part(function_response=function_response)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 0


def test_converter_class_convert_function_call_ignores():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  function_call = genai_types.FunctionCall(
      name=SendA2uiToClientToolset._SendA2uiJsonToClientTool.TOOL_NAME,
      args={SendA2uiToClientToolset._SendA2uiJsonToClientTool.A2UI_JSON_ARG_NAME: "{}"},
  )
  part = genai_types.Part(function_call=function_call)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 0


def test_converter_class_convert_text_with_a2ui():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  valid_a2ui = [{"type": "Text", "text": "Hello"}]
  catalog_mock.validator.validate.return_value = None

  text = f"Here is the UI:\n{A2UI_OPEN_TAG}\n{json.dumps(valid_a2ui)}\n{A2UI_CLOSE_TAG}"
  part = genai_types.Part(text=text)

  a2a_parts = converter.convert(part)

  # Expect 2 parts: TextPart and A2UI DataPart
  assert len(a2a_parts) == 2
  assert a2a_parts[0].root.text == "Here is the UI:"
  assert a2a_parts[1] == create_a2ui_part(valid_a2ui[0])
  catalog_mock.validator.validate.assert_called_once_with(valid_a2ui)


def test_converter_class_convert_text_empty_leading():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  ui = [{"type": "Text", "text": "Top"}]
  catalog_mock.validator.validate.return_value = None

  text = f"\n{A2UI_OPEN_TAG}\n{json.dumps(ui)}\n{A2UI_CLOSE_TAG}"
  part = genai_types.Part(text=text)
  a2a_parts = converter.convert(part)

  assert len(a2a_parts) == 1
  assert a2a_parts[0] == create_a2ui_part(ui[0])


def test_converter_class_convert_text_markdown_wrapped():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  ui = [{"type": "Text", "text": "Inside Markdown"}]
  catalog_mock.validator.validate.return_value = None

  # Text containing JSON wrapped in markdown tags
  text = f"Behold:\n{A2UI_OPEN_TAG}\n```json\n{json.dumps(ui)}\n```\n{A2UI_CLOSE_TAG}"
  part = genai_types.Part(text=text)
  a2a_parts = converter.convert(part)

  assert len(a2a_parts) == 2
  assert a2a_parts[0].root.text == "Behold:"
  assert a2a_parts[1] == create_a2ui_part(ui[0])
  catalog_mock.validator.validate.assert_called_once_with(ui)


def test_converter_class_convert_text_with_invalid_a2ui():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  text = f"Here is the UI:\n{A2UI_OPEN_TAG}\ninvalid_json\n{A2UI_CLOSE_TAG}"
  part = genai_types.Part(text=text)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 0


def test_converter_class_convert_other_part():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  part = genai_types.Part(
      inline_data=genai_types.Blob(mime_type="image/png", data=b"abc")
  )

  with patch(
      "google.adk.a2a.converters.part_converter.convert_genai_part_to_a2a_part"
  ) as mock_convert:
    mock_a2a_part = a2a_types.Part(root=a2a_types.DataPart(kind="data", data={}))
    mock_convert.return_value = mock_a2a_part

    a2a_parts = converter.convert(part)
    assert len(a2a_parts) == 1
    assert a2a_parts[0] is mock_a2a_part
    mock_convert.assert_called_once_with(part)


def test_converter_class_convert_tool_response_with_result_containing_a2ui():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  valid_a2ui = [{"type": "Text", "text": "Result UI"}]
  catalog_mock.validator.validate.return_value = None

  result_text = (
      "Here is the"
      f" result:\n{A2UI_OPEN_TAG}\n{json.dumps(valid_a2ui)}\n{A2UI_CLOSE_TAG}"
  )
  function_response = genai_types.FunctionResponse(
      name="some_generic_tool",
      response={"result": result_text},
  )
  part = genai_types.Part(function_response=function_response)

  a2a_parts = converter.convert(part)

  # Expect 2 parts: TextPart and A2UI DataPart
  assert len(a2a_parts) == 2
  assert a2a_parts[0].root.text == "Here is the result:"
  assert a2a_parts[1] == create_a2ui_part(valid_a2ui[0])
  catalog_mock.validator.validate.assert_called_once_with(valid_a2ui)


def test_converter_class_convert_text_with_invalid_a2ui_and_custom_fallback():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  custom_fallback = "Could not build interface."
  converter = A2uiPartConverter(catalog_mock, fallback_text=custom_fallback)

  text = f"Here is the UI:\n{A2UI_OPEN_TAG}\ninvalid_json\n{A2UI_CLOSE_TAG}"
  part = genai_types.Part(text=text)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 1
  assert a2a_parts[0].root.text == custom_fallback


def test_converter_class_convert_tool_response_with_result_containing_invalid_a2ui_and_default_fallback():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  converter = A2uiPartConverter(catalog_mock)

  result_text = f"Here is the result:\n{A2UI_OPEN_TAG}\ninvalid_json\n{A2UI_CLOSE_TAG}"
  function_response = genai_types.FunctionResponse(
      name="some_generic_tool",
      response={"result": result_text},
  )
  part = genai_types.Part(function_response=function_response)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 0


def test_converter_class_convert_tool_response_with_result_containing_invalid_a2ui_and_custom_fallback():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  custom_fallback = "Could not load the custom tool UI."
  converter = A2uiPartConverter(catalog_mock, fallback_text=custom_fallback)

  result_text = f"Here is the result:\n{A2UI_OPEN_TAG}\ninvalid_json\n{A2UI_CLOSE_TAG}"
  function_response = genai_types.FunctionResponse(
      name="some_generic_tool",
      response={"result": result_text},
  )
  part = genai_types.Part(function_response=function_response)

  a2a_parts = converter.convert(part)
  assert len(a2a_parts) == 1
  assert a2a_parts[0].root.text == custom_fallback



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

from ..schema.catalog import CatalogConfig, resolve_examples_path
from ..schema.catalog_provider import A2uiCatalogProvider
from ..schema.utils import load_from_bundled_resource
from ..schema.constants import BASE_SCHEMA_URL, CATALOG_ID_KEY, CATALOG_SCHEMA_KEY
from .constants import BASIC_CATALOG_NAME, BASIC_CATALOG_PATHS


class BundledCatalogProvider(A2uiCatalogProvider):
  """Loads schemas from bundled package resources with fallbacks."""

  def __init__(self, version: str):
    self.version = version

  def load(self) -> Dict[str, Any]:
    # Use load_from_bundled_resource but with the specialized basic catalog paths
    resource = load_from_bundled_resource(
        self.version, CATALOG_SCHEMA_KEY, BASIC_CATALOG_PATHS
    )

    # Post-load processing for catalogs
    if CATALOG_ID_KEY not in resource:
      spec_map = BASIC_CATALOG_PATHS.get(self.version)
      if spec_map and CATALOG_SCHEMA_KEY in spec_map:
        rel_path = spec_map[CATALOG_SCHEMA_KEY]
        # Strip the `json/` part from the catalog file path for the ID.
        catalog_file = rel_path.replace("/json/", "/")
        resource[CATALOG_ID_KEY] = BASE_SCHEMA_URL + catalog_file

    if "$schema" not in resource:
      resource["$schema"] = "https://json-schema.org/draft/2020-12/schema"

    return resource


class BasicCatalog:
  """Helper for accessing the basic A2UI catalog."""

  @staticmethod
  def get_config(version: str, examples_path: Optional[str] = None) -> CatalogConfig:
    """Returns a CatalogConfig for the basic bundled catalog."""
    return CatalogConfig(
        name=BASIC_CATALOG_NAME,
        provider=BundledCatalogProvider(version),
        examples_path=resolve_examples_path(examples_path),
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the A2UI Event Converter.

This module provides the `A2uiEventConverter` which intercepts ADK events and automatically
translates GenAI model outputs (both tool-based A2UI function calls and text-based delimited A2UI blocks)
into A2A (Agent-to-Agent) event structures with A2UI payloads, using the session catalog if available.

Key Components:
  * `A2uiEventConverter`: An event converter that automatically injects the A2UI catalog into part conversion.

Usage Example:

  Configure the ADK executor to use the A2UI event converter:

  ```python
  config = A2aAgentExecutorConfig(
      event_converter=A2uiEventConverter()
  )
  executor = A2aAgentExecutor(config)
  ```
"""

from typing import TYPE_CHECKING, Optional

from a2ui.adk.a2a.part_converter import A2uiPartConverter
from google.adk.a2a.converters import part_converter
from google.adk.utils.feature_decorator import experimental

if TYPE_CHECKING:
  from a2a.server.events import Event as A2AEvent
  from google.adk.a2a.converters.part_converter import GenAIPartToA2APartConverter
  from google.adk.agents.invocation_context import InvocationContext
  from google.adk.events.event import Event


@experimental
class A2uiEventConverter:
  """An event converter that automatically injects the A2UI catalog into part conversion.

  This allows text-based A2UI extraction and validation to work even when the
  catalog is session-specific.
  """

  def __init__(
      self,
      catalog_key: str = "system:a2ui_catalog",
      bypass_tool_check: bool = False,
      fallback_text: Optional[str] = None,
  ):
    self._catalog_key = catalog_key
    self._bypass_tool_check = bypass_tool_check
    self._fallback_text = fallback_text

  def __call__(
      self,
      event: "Event",
      invocation_context: "InvocationContext",
      task_id: Optional[str] = None,
      context_id: Optional[str] = None,
      part_converter_func: "GenAIPartToA2APartConverter" = part_converter.convert_genai_part_to_a2a_part,
  ) -> list["A2AEvent"]:
    """Converts an ADK event to A2A events, using the session catalog if available."""
    from google.adk.a2a.converters.event_converter import (
        convert_event_to_a2a_events,
    )

    catalog = invocation_context.session.state.get(self._catalog_key)
    if catalog:
      # Use the catalog-aware part converter
      effective_converter = A2uiPartConverter(
          catalog,
          bypass_tool_check=self._bypass_tool_check,
          fallback_text=self._fallback_text,
      ).convert
    else:
      effective_converter = part_converter_func

    return convert_event_to_a2a_events(
        event,
        invocation_context,
        task_id,
        context_id,
        effective_converter,
    )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml
import json
import asyncio
import pytest


def _get_conformance_path(filename):
  return os.path.abspath(
      os.path.join(os.path.dirname(__file__), "../../../conformance", filename)
  )


def load_tests(filename):
  path = _get_conformance_path(os.path.join("suites", filename))
  with open(path, "r", encoding="utf-8") as f:
    return yaml.safe_load(f)


def get_conformance_cases(filename):
  cases = load_tests(filename)
  return [(case["name"], case) for case in cases]


# --- ADK Extensions Conformance ---
cases_adk_extensions = get_conformance_cases("adk_extensions.yaml")


@pytest.mark.parametrize(
    "name, test_case",
    cases_adk_extensions,
    ids=[c[0] for c in cases_adk_extensions],
)
def test_adk_extensions_conformance(name, test_case):
  from a2ui.adk.send_a2ui_to_client_toolset import SendA2uiToClientToolset
  from a2ui.schema.catalog import A2uiCatalog
  from unittest.mock import MagicMock

  action = test_case["action"]
  args = test_case.get("args", {})

  if action == "execute_tool":
    a2ui_json_str = args.get("a2ui_json")
    tool_args = {"a2ui_json": a2ui_json_str} if a2ui_json_str else args

    catalog_mock = MagicMock(spec=A2uiCatalog)
    catalog_mock.validator.validate.return_value = None

    tool = SendA2uiToClientToolset._SendA2uiJsonToClientTool(catalog_mock, "examples")

    tool_context_mock = MagicMock()
    tool_context_mock.state = {}
    tool_context_mock.actions = MagicMock(skip_summarization=False)

    # run_async is async in Python
    result = asyncio.run(tool.run_async(args=tool_args, tool_context=tool_context_mock))

    expect = test_case["expect"]
    expect_success = expect["success"]

    if expect_success:
      assert "error" not in result
      assert (
          SendA2uiToClientToolset._SendA2uiJsonToClientTool.VALIDATED_A2UI_JSON_KEY
          in result
      )
      if expect.get("contains_validated_json"):
        validated_payload = result[
            SendA2uiToClientToolset._SendA2uiJsonToClientTool.VALIDATED_A2UI_JSON_KEY
        ]
        assert "beginRendering" in json.dumps(validated_payload)
    else:
      assert "error" in result
      if expect.get("error_contains"):
        assert expect["error_contains"] in result["error"]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the SendA2uiToClientToolset.

This module provides the necessary components to enable an agent to send A2UI (Agent-to-User Interface)
JSON payloads to a client. It includes the `SendA2uiToClientToolset` for managing A2UI tools, and a specific tool for the LLM
to send JSON.

This is just one approach for capturing A2UI JSON payloads from an LLM.

Key Components:
  * `SendA2uiToClientToolset`: The main entry point. It accepts providers for determining
    if A2UI is enabled and for fetching the A2UI schema. It manages the lifecycle of the
    `_SendA2uiJsonToClientTool`.
  * `_SendA2uiJsonToClientTool`: A tool exposed to the LLM. It allows the LLM to "call" a function
    that effectively sends a JSON payload to the client. This tool validates the JSON against
    the provided schema. It automatically wraps the provided schema in an array structure,
    instructing the LLM that it can send a list of UI items.

Usage Examples:

  1. Defining Providers:
    You can use simple values or callables (sync or async) for enablement, catalog schema, and examples.

    ```python
    # Simple boolean and dict
    toolset = SendA2uiToClientToolset(
        a2ui_enabled=True,
        a2ui_catalog=MY_CATALOG,
        a2ui_examples=MY_EXAMPLES,
    )

    # Async providers
    async def check_enabled(ctx: ReadonlyContext) -> bool:
      return await some_condition(ctx)

    async def get_catalog(ctx: ReadonlyContext) -> catalog.A2uiCatalog:
      return await fetch_catalog(ctx)

    async def get_examples(ctx: ReadonlyContext) -> str:
      return await fetch_examples(ctx)

    toolset = SendA2uiToClientToolset(
        a2ui_enabled=check_enabled,
        a2ui_catalog=get_catalog,
        a2ui_examples=get_examples,
    )
    ```

  2. Integration with Agent:
    Typically used when initializing an agent's toolset.

    ```python
    # In your agent initialization
    LlmAgent(
        tools=[
            SendA2uiToClientToolset(
                a2ui_enabled=check_enabled,
                a2ui_catalog=get_catalog,
                a2ui_examples=get_examples,
            ),
        ],
    )
    ```
"""

import inspect
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Optional,
    TypeAlias,
    Union,
)

from a2ui.adk.a2a.event_converter import A2uiEventConverter
from a2ui.adk.a2a.part_converter import A2uiPartConverter
from a2ui.parser.payload_fixer import parse_and_fix
from a2ui.schema import catalog
from a2ui.schema import constants
from a2ui.schema.catalog import A2uiCatalog
from a2ui.schema.constants import (
    A2UI_SCHEMA_BLOCK_END,
    A2UI_SCHEMA_BLOCK_START,
    A2UI_TOOL_ERROR_KEY,
    A2UI_TOOL_NAME,
    A2UI_VALIDATED_JSON_KEY,
)
from google.adk.agents import readonly_context
from google.adk.agents.readonly_context import ReadonlyContext
from google.adk.models import LlmRequest
from google.adk.tools import base_toolset
from google.adk.tools import base_tool
from google.adk.tools import tool_context
from google.adk.utils.feature_decorator import experimental
from google.genai import types as genai_types

if TYPE_CHECKING:
  from google.adk.agents.invocation_context import InvocationContext
  from google.adk.events.event import Event

logger = logging.getLogger(__name__)

A2uiEnabledProvider: TypeAlias = Callable[
    [readonly_context.ReadonlyContext], Union[bool, Awaitable[bool]]
]
A2uiCatalogProvider: TypeAlias = Callable[
    [readonly_context.ReadonlyContext],
    Union[catalog.A2uiCatalog, Awaitable[catalog.A2uiCatalog]],
]
A2uiExamplesProvider: TypeAlias = Callable[
    [readonly_context.ReadonlyContext], Union[str, Awaitable[str]]
]


@experimental
class SendA2uiToClientToolset(base_toolset.BaseToolset):
  """A toolset that provides A2UI Tools and can be enabled/disabled."""

  def __init__(
      self,
      a2ui_enabled: Union[bool, A2uiEnabledProvider],
      a2ui_catalog: Union[catalog.A2uiCatalog, A2uiCatalogProvider],
      a2ui_examples: Union[str, A2uiExamplesProvider],
  ):
    super().__init__()
    self._a2ui_enabled = a2ui_enabled
    self._ui_tools = [self._SendA2uiJsonToClientTool(a2ui_catalog, a2ui_examples)]

  async def _resolve_a2ui_enabled(self, ctx: readonly_context.ReadonlyContext) -> bool:
    """The resolved self.a2ui_enabled field to construct instruction for this agent.

    Args:
        ctx: The ReadonlyContext to resolve the provider with.

    Returns:
        If A2UI is enabled, return True. Otherwise, return False.
    """
    if isinstance(self._a2ui_enabled, bool):
      return self._a2ui_enabled
    else:
      a2ui_enabled = self._a2ui_enabled(ctx)
      if inspect.isawaitable(a2ui_enabled):
        a2ui_enabled = await a2ui_enabled
    return a2ui_enabled

  async def get_tools(
      self,
      readonly_context: Optional[readonly_context.ReadonlyContext] = None,
  ) -> list[base_tool.BaseTool]:
    """Returns the list of tools provided by this toolset.

    Args:
        readonly_context: The ReadonlyContext for resolving tool enablement.

    Returns:
        A list of tools.
    """
    use_ui = False
    if readonly_context is not None:
      use_ui = await self._resolve_a2ui_enabled(readonly_context)
    if use_ui:
      logger.info("A2UI is ENABLED, adding ui tools")
      return self._ui_tools
    else:
      logger.info("A2UI is DISABLED, not adding ui tools")
      return []

  async def get_part_converter(
      self, ctx: readonly_context.ReadonlyContext
  ) -> "A2uiPartConverter":
    """Returns a configured A2uiPartConverter for the given context.

    Args:
        ctx: The readonly_context.ReadonlyContext to resolve the catalog with.

    Returns:
        A configured A2uiPartConverter.
    """
    catalog = await self._ui_tools[0]._resolve_a2ui_catalog(ctx)
    return A2uiPartConverter(catalog)

  class _SendA2uiJsonToClientTool(base_tool.BaseTool):
    TOOL_NAME = A2UI_TOOL_NAME
    VALIDATED_A2UI_JSON_KEY = A2UI_VALIDATED_JSON_KEY
    A2UI_JSON_ARG_NAME = "a2ui_json"
    TOOL_ERROR_KEY = A2UI_TOOL_ERROR_KEY

    def __init__(
        self,
        a2ui_catalog: Union[catalog.A2uiCatalog, A2uiCatalogProvider],
        a2ui_examples: Union[str, A2uiExamplesProvider],
    ):
      self._a2ui_catalog = a2ui_catalog
      self._a2ui_examples = a2ui_examples
      super().__init__(
          name=self.TOOL_NAME,
          description=(
              "Sends A2UI JSON to the client to render rich UI for the user. This tool"
              " can be called multiple times in the same call to render multiple UI"
              f" surfaces.Args:    {self.A2UI_JSON_ARG_NAME}: Valid A2UI JSON Schema to"
              " send to the client. The A2UI JSON Schema definition is between"
              f" {constants.A2UI_SCHEMA_BLOCK_START} and"
              f" {constants.A2UI_SCHEMA_BLOCK_END} in the system instructions."
          ),
      )

    def _get_declaration(self) -> genai_types.FunctionDeclaration | None:
      return genai_types.FunctionDeclaration(
          name=self.name,
          description=self.description,
          parameters=genai_types.Schema(
              type=genai_types.Type.OBJECT,
              properties={
                  self.A2UI_JSON_ARG_NAME: genai_types.Schema(
                      type=genai_types.Type.STRING,
                      description="valid A2UI JSON Schema to send to the client.",
                  ),
              },
              required=[self.A2UI_JSON_ARG_NAME],
          ),
      )

    async def _resolve_a2ui_examples(
        self, ctx: readonly_context.ReadonlyContext
    ) -> str:
      """The resolved self.a2ui_examples field to construct instruction for this agent.

      Args:
          ctx: The readonly_context.ReadonlyContext to resolve the provider with.

      Returns:
          The A2UI examples string.
      """
      if isinstance(self._a2ui_examples, str):
        return self._a2ui_examples
      else:
        a2ui_examples = self._a2ui_examples(ctx)
        if inspect.isawaitable(a2ui_examples):
          a2ui_examples = await a2ui_examples
        return a2ui_examples

    async def _resolve_a2ui_catalog(
        self, ctx: readonly_context.ReadonlyContext
    ) -> catalog.A2uiCatalog:
      """The resolved self.a2ui_catalog field to construct instruction for this agent.

      Args:
          ctx: The readonly_context.ReadonlyContext to resolve the provider with.

      Returns:
          The A2UI catalog object.
      """
      if isinstance(self._a2ui_catalog, catalog.A2uiCatalog):
        return self._a2ui_catalog
      else:
        a2ui_catalog = self._a2ui_catalog(ctx)
        if inspect.isawaitable(a2ui_catalog):
          a2ui_catalog = await a2ui_catalog
        return a2ui_catalog

    async def process_llm_request(
        self,
        *,
        tool_context: tool_context.ToolContext,
        llm_request: models.LlmRequest,
    ) -> None:
      await super().process_llm_request(
          tool_context=tool_context, llm_request=llm_request
      )

      a2ui_catalog = await self._resolve_a2ui_catalog(tool_context)

      instruction = a2ui_catalog.render_as_llm_instructions()
      examples = await self._resolve_a2ui_examples(tool_context)

      llm_request.append_instructions([instruction, examples])

      logger.info("Added A2UI schema and examples to system instructions")

    async def run_async(
        self, *, args: dict[str, Any], tool_context: tool_context.ToolContext
    ) -> Any:
      try:
        a2ui_json = args.get(self.A2UI_JSON_ARG_NAME)
        if not a2ui_json:
          raise ValueError(
              f"Failed to call tool {self.TOOL_NAME} because missing required"
              f" arg {self.A2UI_JSON_ARG_NAME} "
          )

        a2ui_catalog = await self._resolve_a2ui_catalog(tool_context)
        a2ui_json_payload = parse_and_fix(a2ui_json)
        a2ui_catalog.validator.validate(a2ui_json_payload)

        logger.info(
            f"Validated call to tool {self.TOOL_NAME} with {self.A2UI_JSON_ARG_NAME}"
        )

        # Don't do a second LLM inference call for the JSON response
        tool_context.actions.skip_summarization = True

        # Return the validated JSON so the converter can use it.
        # We return it in a dict under "result" key for consistent JSON structure.
        return {self.VALIDATED_A2UI_JSON_KEY: a2ui_json_payload}

      except Exception as e:
        err = f"Failed to call A2UI tool {self.TOOL_NAME}: {e}"
        logger.error(err)

        return {self.TOOL_ERROR_KEY: err}



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from a2ui.schema.manager import A2uiSchemaManager
from a2ui.schema.constants import CATALOG_COMPONENTS_KEY, VERSION_0_8, VERSION_0_9
from a2ui.schema.common_modifiers import remove_strict_validation
from a2ui.basic_catalog.provider import BasicCatalog


def verify():
  print('Verifying A2uiSchemaManager...')
  try:
    manager = A2uiSchemaManager(
        version=VERSION_0_8,
        catalogs=[BasicCatalog.get_config(VERSION_0_8)],
        schema_modifiers=[remove_strict_validation],
    )
    catalog = manager.get_selected_catalog()
    catalog_components = catalog.catalog_schema[CATALOG_COMPONENTS_KEY]
    print(f'Successfully loaded {VERSION_0_8}: {len(catalog_components)} components')
    print(f'Components found: {list(catalog_components.keys())[:5]}...')

    a2ui_message = [
        {'beginRendering': {'surfaceId': 'contact-card', 'root': 'main_card'}},
        {
            'surfaceUpdate': {
                'surfaceId': 'contact-card',
                'components': [
                    {
                        'id': 'profile_image',
                        'component': {
                            'Image': {
                                'url': {'path': '/imageUrl'},
                                'usageHint': 'avatar',
                                'fit': 'cover',
                            }
                        },
                    },
                    {
                        'id': 'user_heading',
                        'weight': 1,
                        'component': {
                            'Text': {'text': {'path': '/name'}, 'usageHint': 'h2'}
                        },
                    },
                    {
                        'id': 'description_text_1',
                        'component': {'Text': {'text': {'path': '/title'}}},
                    },
                    {
                        'id': 'description_text_2',
                        'component': {'Text': {'text': {'path': '/team'}}},
                    },
                    {
                        'id': 'description_column',
                        'component': {
                            'Column': {
                                'children': {
                                    'explicitList': [
                                        'user_heading',
                                        'description_text_1',
                                        'description_text_2',
                                    ]
                                },
                                'alignment': 'center',
                            }
                        },
                    },
                    {
                        'id': 'calendar_icon',
                        'component': {
                            'Icon': {'name': {'literalString': 'calendarToday'}}
                        },
                    },
                    {
                        'id': 'calendar_primary_text',
                        'component': {
                            'Text': {'usageHint': 'h5', 'text': {'path': '/calendar'}}
                        },
                    },
                    {
                        'id': 'calendar_secondary_text',
                        'component': {'Text': {'text': {'literalString': 'Calendar'}}},
                    },
                    {
                        'id': 'calendar_text_column',
                        'component': {
                            'Column': {
                                'children': {
                                    'explicitList': [
                                        'calendar_primary_text',
                                        'calendar_secondary_text',
                                    ]
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {
                        'id': 'info_row_1',
                        'component': {
                            'Row': {
                                'children': {
                                    'explicitList': [
                                        'calendar_icon',
                                        'calendar_text_column',
                                    ]
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {
                        'id': 'location_icon',
                        'component': {
                            'Icon': {'name': {'literalString': 'locationOn'}}
                        },
                    },
                    {
                        'id': 'location_primary_text',
                        'component': {
                            'Text': {'usageHint': 'h5', 'text': {'path': '/location'}}
                        },
                    },
                    {
                        'id': 'location_secondary_text',
                        'component': {'Text': {'text': {'literalString': 'Location'}}},
                    },
                    {
                        'id': 'location_text_column',
                        'component': {
                            'Column': {
                                'children': {
                                    'explicitList': [
                                        'location_primary_text',
                                        'location_secondary_text',
                                    ]
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {
                        'id': 'info_row_2',
                        'component': {
                            'Row': {
                                'children': {
                                    'explicitList': [
                                        'location_icon',
                                        'location_text_column',
                                    ]
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {
                        'id': 'mail_icon',
                        'component': {'Icon': {'name': {'literalString': 'mail'}}},
                    },
                    {
                        'id': 'mail_primary_text',
                        'component': {
                            'Text': {'usageHint': 'h5', 'text': {'path': '/email'}}
                        },
                    },
                    {
                        'id': 'mail_secondary_text',
                        'component': {'Text': {'text': {'literalString': 'Email'}}},
                    },
                    {
                        'id': 'mail_text_column',
                        'component': {
                            'Column': {
                                'children': {
                                    'explicitList': [
                                        'mail_primary_text',
                                        'mail_secondary_text',
                                    ]
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {
                        'id': 'info_row_3',
                        'component': {
                            'Row': {
                                'children': {
                                    'explicitList': ['mail_icon', 'mail_text_column']
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {'id': 'div', 'component': {'Divider': {}}},
                    {
                        'id': 'call_icon',
                        'component': {'Icon': {'name': {'literalString': 'call'}}},
                    },
                    {
                        'id': 'call_primary_text',
                        'component': {
                            'Text': {'usageHint': 'h5', 'text': {'path': '/mobile'}}
                        },
                    },
                    {
                        'id': 'call_secondary_text',
                        'component': {'Text': {'text': {'literalString': 'Mobile'}}},
                    },
                    {
                        'id': 'call_text_column',
                        'component': {
                            'Column': {
                                'children': {
                                    'explicitList': [
                                        'call_primary_text',
                                        'call_secondary_text',
                                    ]
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {
                        'id': 'info_row_4',
                        'component': {
                            'Row': {
                                'children': {
                                    'explicitList': ['call_icon', 'call_text_column']
                                },
                                'distribution': 'start',
                                'alignment': 'start',
                            }
                        },
                    },
                    {
                        'id': 'info_rows_column',
                        'weight': 1,
                        'component': {
                            'Column': {
                                'children': {
                                    'explicitList': [
                                        'info_row_1',
                                        'info_row_2',
                                        'info_row_3',
                                        'info_row_4',
                                    ]
                                },
                                'alignment': 'stretch',
                            }
                        },
                    },
                    {
                        'id': 'button_1_text',
                        'component': {'Text': {'text': {'literalString': 'Follow'}}},
                    },
                    {
                        'id': 'button_1',
                        'component': {
                            'Button': {
                                'child': 'button_1_text',
                                'primary': True,
                                'action': {'name': 'follow_contact'},
                            }
                        },
                    },
                    {
                        'id': 'button_2_text',
                        'component': {'Text': {'text': {'literalString': 'Message'}}},
                    },
                    {
                        'id': 'button_2',
                        'component': {
                            'Button': {
                                'child': 'button_2_text',
                                'primary': False,
                                'action': {'name': 'send_message'},
                            }
                        },
                    },
                    {
                        'id': 'action_buttons_row',
                        'component': {
                            'Row': {
                                'children': {'explicitList': ['button_1', 'button_2']},
                                'distribution': 'center',
                                'alignment': 'center',
                            }
                        },
                    },
                    {
                        'id': 'link_text',
                        'component': {
                            'Text': {
                                'text': {
                                    'literalString': '[View Full Profile](/profile)'
                                }
                            }
                        },
                    },
                    {
                        'id': 'link_text_wrapper',
                        'component': {
                            'Row': {
                                'children': {'explicitList': ['link_text']},
                                'distribution': 'center',
                                'alignment': 'center',
                            }
                        },
                    },
                    {
                        'id': 'main_column',
                        'component': {
                            'Column': {
                                'children': {
                                    'explicitList': [
                                        'profile_image',
                                        'description_column',
                                        'div',
                                        'info_rows_column',
                                        'action_buttons_row',
                                        'link_text_wrapper',
                                    ]
                                },
                                'alignment': 'stretch',
                            }
                        },
                    },
                    {
                        'id': 'main_card',
                        'component': {'Card': {'child': 'main_column'}},
                    },
                ],
            }
        },
        {
            'dataModelUpdate': {
                'surfaceId': 'contact-card',
                'path': '/',
                'contents': [
                    {'key': 'name', 'valueString': 'Casey Smith'},
                    {'key': 'title', 'valueString': 'Digital Marketing Specialist'},
                    {'key': 'team', 'valueString': 'Growth Team'},
                    {'key': 'location', 'valueString': 'New York'},
                    {'key': 'email', 'valueString': 'casey.smith@example.com'},
                    {'key': 'mobile', 'valueString': '+1 (415) 222-3333'},
                    {'key': 'calendar', 'valueString': 'In a meeting'},
                    {
                        'key': 'imageUrl',
                        'valueString': 'http://localhost:10003/static/profile2.png',
                    },
                    {
                        'key': 'contacts',
                        'valueMap': [{
                            'key': 'contact1',
                            'valueMap': [{'key': 'name', 'valueString': 'Casey Smith'}],
                        }],
                    },
                ],
            }
        },
    ]
    catalog.validator.validate(a2ui_message)
    print('Validation successful')
  except Exception as e:
    print(f'Failed to load {VERSION_0_8}: {e}')
    sys.exit(1)

  try:
    manager = A2uiSchemaManager(
        version=VERSION_0_9,
        catalogs=[BasicCatalog.get_config(VERSION_0_9)],
        schema_modifiers=[remove_strict_validation],
    )
    catalog = manager.get_selected_catalog()
    catalog_components = catalog.catalog_schema[CATALOG_COMPONENTS_KEY]
    print(f'Successfully loaded {VERSION_0_9}: {len(catalog_components)} components')
    print(f'Components found: {list(catalog_components.keys())}...')

    a2ui_message = [
        {
            'version': 'v0.9',
            'createSurface': {
                'surfaceId': 'contact_form_1',
                'catalogId': (
                    'https://a2ui.dev/specification/v0_9/catalogs/basic/catalog.json'
                ),
                'fakeProperty': 'should be allowed',
            },
        },
        {
            'version': 'v0.9',
            'updateComponents': {
                'surfaceId': 'contact_form_1',
                'components': [
                    {'id': 'root', 'component': 'Card', 'child': 'form_container'},
                    {
                        'id': 'form_container',
                        'component': 'Column',
                        'children': [
                            'header_row',
                            'name_row',
                            'email_group',
                            'phone_group',
                            'pref_group',
                            'divider_1',
                            'newsletter_checkbox',
                            'submit_button',
                        ],
                        'justify': 'start',
                        'align': 'stretch',
                    },
                    {
                        'id': 'header_row',
                        'component': 'Row',
                        'children': ['header_icon', 'header_text'],
                        'align': 'center',
                    },
                    {'id': 'header_icon', 'component': 'Icon', 'name': 'mail'},
                    {
                        'id': 'header_text',
                        'component': 'Text',
                        'text': '# Contact Us',
                        'variant': 'h2',
                    },
                    {
                        'id': 'name_row',
                        'component': 'Row',
                        'children': ['first_name_group', 'last_name_group'],
                        'justify': 'spaceBetween',
                    },
                    {
                        'id': 'first_name_group',
                        'component': 'Column',
                        'children': ['first_name_label', 'first_name_field'],
                        'weight': 1,
                    },
                    {
                        'id': 'first_name_label',
                        'component': 'Text',
                        'text': 'First Name',
                        'variant': 'caption',
                    },
                    {
                        'id': 'first_name_field',
                        'component': 'TextField',
                        'label': 'First Name',
                        'value': {'path': '/contact/firstName'},
                        'variant': 'shortText',
                    },
                    {
                        'id': 'last_name_group',
                        'component': 'Column',
                        'children': ['last_name_label', 'last_name_field'],
                        'weight': 1,
                    },
                    {
                        'id': 'last_name_label',
                        'component': 'Text',
                        'text': 'Last Name',
                        'variant': 'caption',
                    },
                    {
                        'id': 'last_name_field',
                        'component': 'TextField',
                        'label': 'Last Name',
                        'value': {'path': '/contact/lastName'},
                        'variant': 'shortText',
                    },
                    {
                        'id': 'email_group',
                        'component': 'Column',
                        'children': ['email_label', 'email_field'],
                    },
                    {
                        'id': 'email_label',
                        'component': 'Text',
                        'text': 'Email Address',
                        'variant': 'caption',
                    },
                    {
                        'id': 'email_field',
                        'component': 'TextField',
                        'label': 'Email',
                        'value': {'path': '/contact/email'},
                        'variant': 'shortText',
                        'checks': [
                            {
                                'condition': {
                                    'call': 'required',
                                    'args': {'value': {'path': '/contact/email'}},
                                },
                                'message': 'Email is required.',
                            },
                            {
                                'condition': {
                                    'call': 'email',
                                    'args': {'value': {'path': '/contact/email'}},
                                },
                                'message': 'Please enter a valid email address.',
                            },
                        ],
                    },
                    {
                        'id': 'phone_group',
                        'component': 'Column',
                        'children': ['phone_label', 'phone_field'],
                    },
                    {
                        'id': 'phone_label',
                        'component': 'Text',
                        'text': 'Phone Number',
                        'variant': 'caption',
                    },
                    {
                        'id': 'phone_field',
                        'component': 'TextField',
                        'label': 'Phone',
                        'value': {'path': '/contact/phone'},
                        'variant': 'shortText',
                        'checks': [{
                            'condition': {
                                'call': 'regex',
                                'args': {
                                    'value': {'path': '/contact/phone'},
                                    'pattern': '^\\d{10}$',
                                },
                            },
                            'message': 'Phone number must be 10 digits.',
                        }],
                    },
                    {
                        'id': 'pref_group',
                        'component': 'Column',
                        'children': ['pref_label', 'pref_picker'],
                    },
                    {
                        'id': 'pref_label',
                        'component': 'Text',
                        'text': 'Preferred Contact Method',
                        'variant': 'caption',
                    },
                    {
                        'id': 'pref_picker',
                        'component': 'ChoicePicker',
                        'variant': 'mutuallyExclusive',
                        'options': [
                            {'label': 'Email', 'value': 'email'},
                            {'label': 'Phone', 'value': 'phone'},
                            {'label': 'SMS', 'value': 'sms'},
                        ],
                        'value': {'path': '/contact/preference'},
                    },
                    {'id': 'divider_1', 'component': 'Divider', 'axis': 'horizontal'},
                    {
                        'id': 'newsletter_checkbox',
                        'component': 'CheckBox',
                        'label': 'Subscribe to our newsletter',
                        'value': {'path': '/contact/subscribe'},
                    },
                    {
                        'id': 'submit_button_label',
                        'component': 'Text',
                        'text': 'Send Message',
                    },
                    {
                        'id': 'submit_button',
                        'component': 'Button',
                        'child': 'submit_button_label',
                        'variant': 'primary',
                        'action': {
                            'event': {
                                'name': 'submitContactForm',
                                'context': {
                                    'formId': 'contact_form_1',
                                    'isNewsletterSubscribed': {
                                        'path': '/contact/subscribe'
                                    },
                                },
                            }
                        },
                    },
                ],
            },
        },
        {
            'version': 'v0.9',
            'updateDataModel': {
                'surfaceId': 'contact_form_1',
                'path': '/contact',
                'value': {
                    'firstName': 'John',
                    'lastName': 'Doe',
                    'email': 'john.doe@example.com',
                    'phone': '1234567890',
                    'preference': ['email'],
                    'subscribe': True,
                },
            },
        },
        {'version': 'v0.9', 'deleteSurface': {'surfaceId': 'contact_form_1'}},
    ]
    catalog.validator.validate(a2ui_message)
    print('Validation successful')
  except Exception as e:
    print(f'Failed to load {VERSION_0_9}: {e}')
    sys.exit(1)


if __name__ == '__main__':
  verify()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml
import pytest

from a2ui.basic_catalog import BasicCatalog
from a2ui.schema.catalog import A2uiCatalog
from a2ui.parser.streaming import A2uiStreamParser
from a2ui.schema.validator import A2uiValidator
from a2ui.schema.manager import A2uiSchemaManager, CatalogConfig
from a2ui.schema.common_modifiers import remove_strict_validation
from a2ui.schema.constants import VERSION_0_8, VERSION_0_9

import json
import re


class MemoryCatalogProvider:

  def __init__(self, schema):
    self.schema = schema

  def load(self):
    return self.schema


def _get_conformance_path(filename):
  return os.path.abspath(
      os.path.join(os.path.dirname(__file__), "../../../conformance", filename)
  )


def load_json_file(filename):
  path = _get_conformance_path(filename)
  with open(path, "r", encoding="utf-8") as f:
    return json.load(f)


def load_tests(filename):
  path = _get_conformance_path(os.path.join("suites", filename))
  with open(path, "r", encoding="utf-8") as f:

    return yaml.safe_load(f)


def setup_catalog(catalog_config):
  version = str(catalog_config["version"])

  s2c_schema = catalog_config.get("s2c_schema")
  if isinstance(s2c_schema, str):
    s2c_schema = load_json_file(s2c_schema)

  catalog_schema = catalog_config.get("catalog_schema")
  if isinstance(catalog_schema, str):
    catalog_schema = load_json_file(catalog_schema)
  elif catalog_schema is None:
    catalog_schema = {}

  common_types_schema = catalog_config.get("common_types_schema")
  if isinstance(common_types_schema, str):
    common_types_schema = load_json_file(common_types_schema)
  elif common_types_schema is None:
    common_types_schema = {}

  return A2uiCatalog(
      version=version,
      name="test_catalog",
      s2c_schema=s2c_schema,
      common_types_schema=common_types_schema,
      catalog_schema=catalog_schema,
  )


def assert_parts_match(actual_parts, expected_parts):
  assert len(actual_parts) == len(expected_parts)
  for actual, expected in zip(actual_parts, expected_parts):
    assert actual.text == expected.get("text", "")
    assert actual.a2ui_json == expected.get("a2ui")


def get_conformance_cases(filename):
  cases = load_tests(filename)
  return [(case["name"], case) for case in cases]


# --- Streaming Parser Conformance ---
cases_parser = get_conformance_cases("streaming_parser.yaml")


@pytest.mark.parametrize(
    "name, test_case", cases_parser, ids=[c[0] for c in cases_parser]
)
def test_parser_conformance(name, test_case):
  catalog_config = test_case["catalog"]
  catalog = setup_catalog(catalog_config)
  parser = A2uiStreamParser(catalog=catalog)

  steps = test_case.get("steps")
  if steps is None and "process_chunk" in test_case:
    steps = test_case["process_chunk"]

  if steps is None and "input" in test_case:
    steps = [test_case]

  for step in steps:
    expect_error = step.get("expect_error") or test_case.get("expect_error")
    if expect_error:
      with pytest.raises(ValueError, match=expect_error):
        parser.process_chunk(step["input"])
    else:
      parts = parser.process_chunk(step["input"])
      assert_parts_match(parts, step["expect"])


# --- Non-Streaming Parser Conformance ---
cases_parser_non_streaming = get_conformance_cases("parser.yaml")


@pytest.mark.parametrize(
    "name, test_case",
    cases_parser_non_streaming,
    ids=[c[0] for c in cases_parser_non_streaming],
)
def test_parser_non_streaming_conformance(name, test_case):
  from a2ui.parser.parser import parse_response
  from a2ui.parser.payload_fixer import parse_and_fix

  action = test_case.get("action", "parse_full")
  content = test_case["input"]

  if action == "parse_full":
    if "expect_error" in test_case:
      with pytest.raises(ValueError, match=test_case["expect_error"]):
        parse_response(content)
    else:
      parts = parse_response(content)
      expected = test_case["expect"]
      assert len(parts) == len(expected)
      for actual, exp in zip(parts, expected):
        assert actual.text.strip() == exp.get("text", "").strip()
        assert actual.a2ui_json == exp.get("a2ui")

  elif action == "fix_payload":
    if "expect_error" in test_case:
      with pytest.raises(ValueError, match=test_case["expect_error"]):
        parse_and_fix(content)
    else:
      result = parse_and_fix(content)
      assert result == test_case["expect"]

  elif action == "has_parts":
    from a2ui.parser.parser import has_a2ui_parts

    result = has_a2ui_parts(content)
    assert result == test_case["expect"]


# --- Validator Conformance ---

cases_validator = get_conformance_cases("validator.yaml")


@pytest.mark.parametrize(
    "name, test_case", cases_validator, ids=[c[0] for c in cases_validator]
)
def test_validator_conformance(name, test_case):
  catalog_config = test_case["catalog"]
  catalog = setup_catalog(catalog_config)

  steps = test_case.get("steps")
  if steps is None and "validate" in test_case:
    steps = test_case["validate"]

  if steps is None and "payload" in test_case:
    steps = [test_case]

  for step in steps:
    validator = A2uiValidator(catalog=catalog)
    expect_error = step.get("expect_error") or test_case.get("expect_error")
    if expect_error:
      with pytest.raises(ValueError, match=expect_error):
        validator.validate(step["payload"])
    else:
      validator.validate(step["payload"])


# --- Catalog Conformance ---
cases_catalog = get_conformance_cases("catalog.yaml")


@pytest.mark.parametrize(
    "name, test_case", cases_catalog, ids=[c[0] for c in cases_catalog]
)
def test_catalog_conformance(name, test_case):
  catalog_config = test_case["catalog"]
  catalog = setup_catalog(catalog_config)
  action = test_case["action"]
  args = test_case.get("args", {})

  if action == "prune":
    allowed_components = args.get("allowed_components", [])
    allowed_messages = args.get("allowed_messages", [])
    pruned = catalog.with_pruning(allowed_components, allowed_messages)
    expected = test_case["expect"]
    if "catalog_schema" in expected:
      assert pruned.catalog_schema == expected["catalog_schema"]
    if "s2c_schema" in expected:
      assert pruned.s2c_schema == expected["s2c_schema"]
    if "common_types_schema" in expected:
      assert pruned.common_types_schema == expected["common_types_schema"]

  elif action == "render":
    output = catalog.render_as_llm_instructions()
    assert output.strip() == test_case["expect_output"].strip()

  elif action == "load":
    path = args.get("path")
    if path:
      full_path = os.path.join(os.path.dirname(__file__), "../../../conformance", path)
    else:
      full_path = None
    validate = args.get("validate", False)
    if "expect_error" in test_case:
      with pytest.raises(ValueError, match=test_case["expect_error"]):
        catalog.load_examples(full_path, validate=validate)
    else:
      output = catalog.load_examples(full_path, validate=validate)
      assert output.strip() == test_case["expect_output"].strip()

  elif action == "remove_strict_validation":
    schema = args["schema"]
    modified = remove_strict_validation(schema)
    assert modified == test_case["expect"]["schema"]


# --- Schema Manager Conformance ---
cases_schema_manager = get_conformance_cases("schema_manager.yaml")


@pytest.mark.parametrize(
    "name, test_case",
    cases_schema_manager,
    ids=[c[0] for c in cases_schema_manager],
)
def test_schema_manager_conformance(name, test_case):
  action = test_case["action"]
  args = test_case.get("args", {})

  if action == "select_catalog":
    supported_catalogs = args.get("supported_catalogs", [])
    client_capabilities = args.get("client_capabilities", {})
    accepts_inline_catalogs = args.get("accepts_inline_catalogs", False)

    configs = []
    for cat_def in supported_catalogs:
      configs.append(
          CatalogConfig(
              name=cat_def["catalogId"],
              provider=MemoryCatalogProvider(cat_def),
          )
      )

    manager = A2uiSchemaManager(
        version=VERSION_0_9,
        catalogs=configs,
        accepts_inline_catalogs=accepts_inline_catalogs,
    )

    if "expect_error" in test_case:
      with pytest.raises(ValueError, match=test_case["expect_error"]):
        manager.get_selected_catalog(client_capabilities)
    else:
      selected = manager.get_selected_catalog(client_capabilities)
      if "expect_selected" in test_case:
        assert selected.catalog_id == test_case["expect_selected"]
      if "expect_catalog_schema" in test_case:
        assert selected.catalog_schema == test_case["expect_catalog_schema"]

  elif action == "load_catalog":
    catalog_configs = test_case.get("catalog_configs", [])
    modifiers = test_case.get("modifiers", [])
    schema_modifiers = []
    if "remove_strict_validation" in modifiers:
      schema_modifiers.append(remove_strict_validation)
    configs = []
    for cfg in catalog_configs:
      full_path = os.path.join(
          os.path.dirname(__file__), "../../../conformance", cfg["path"]
      )
      configs.append(CatalogConfig.from_path(name=cfg["name"], catalog_path=full_path))
    manager = A2uiSchemaManager(
        version=VERSION_0_8, catalogs=configs, schema_modifiers=schema_modifiers
    )
    selected = manager.get_selected_catalog()
    expected = test_case["expect"]
    if "catalog_schema" in expected:
      assert selected.catalog_schema == expected["catalog_schema"]
    if "supported_catalog_ids" in expected:
      assert [c.catalog_id for c in manager._supported_catalogs] == expected[
          "supported_catalog_ids"
      ]

  elif action == "generate_prompt":
    version = args.get("version", VERSION_0_8)
    role = args.get("role_description", "")
    workflow = args.get("workflow_description", "")
    ui_desc = args.get("ui_description", "")

    examples_path = args.get("examples_path")
    if examples_path:
      examples_path = os.path.join(
          os.path.dirname(__file__), "../../../conformance", examples_path
      )

    config = BasicCatalog.get_config(version)
    if examples_path:
      config = CatalogConfig(
          name=config.name,
          provider=config.provider,
          examples_path=examples_path,
      )

    manager = A2uiSchemaManager(
        version=version,
        catalogs=[config],
        accepts_inline_catalogs=args.get("accepts_inline_catalogs", False),
    )

    output = manager.generate_system_prompt(
        role_description=role,
        workflow_description=workflow,
        ui_description=ui_desc,
        include_schema=args.get("include_schema", False),
        include_examples=args.get("include_examples", False),
        client_ui_capabilities=args.get("client_ui_capabilities"),
        allowed_components=args.get("allowed_components"),
        allowed_messages=args.get("allowed_messages"),
    )

    output_normalized = re.sub(r"\s+", "", output.strip())

    if "expect_contains" in test_case:
      for expected in test_case["expect_contains"]:
        expected_normalized = re.sub(r"\s+", "", expected.strip())
        assert expected_normalized in output_normalized



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import io
import pytest
import json
import os
from unittest.mock import patch, MagicMock, PropertyMock
from a2ui.schema.manager import A2uiSchemaManager, A2uiCatalog, CatalogConfig
from a2ui.basic_catalog import BasicCatalog
from a2ui.basic_catalog.constants import BASIC_CATALOG_NAME
from a2ui.schema.constants import (
    DEFAULT_WORKFLOW_RULES,
    INLINE_CATALOG_NAME,
    VERSION_0_8,
    VERSION_0_9,
)
from a2ui.schema.constants import (
    A2UI_SCHEMA_BLOCK_START,
    A2UI_SCHEMA_BLOCK_END,
    INLINE_CATALOGS_KEY,
    SUPPORTED_CATALOG_IDS_KEY,
)


@pytest.fixture
def mock_importlib_resources():
  with patch("importlib.resources.files") as mock_files:
    yield mock_files


def test_schema_manager_init_valid_version(mock_importlib_resources):
  mock_files = mock_importlib_resources
  mock_traversable = MagicMock()

  def files_side_effect(package):
    if package == "a2ui.assets":
      return mock_traversable
    return MagicMock()

  mock_files.side_effect = files_side_effect

  # Mock file open calls for server_to_client and catalog
  def joinpath_side_effect(path):
    if path == VERSION_0_8:
      return mock_traversable

    mock_file = MagicMock()
    if path == "server_to_client.json":
      content = (
          '{"$schema": "https://json-schema.org/draft/2020-12/schema", "version":'
          f' "{VERSION_0_8}", "defs": "server_defs"}}'
      )
    elif path == "standard_catalog_definition.json":
      content = (
          '{"$schema": "https://json-schema.org/draft/2020-12/schema", "version":'
          f' "{VERSION_0_8}", "components": {{"Text": {{}}}}}}'
      )
    else:
      content = '{"$schema": "https://json-schema.org/draft/2020-12/schema"}'

    mock_file.open.return_value.__enter__.return_value = io.StringIO(content)
    return mock_file

  mock_traversable.joinpath.side_effect = joinpath_side_effect

  manager = A2uiSchemaManager(
      VERSION_0_8, catalogs=[BasicCatalog.get_config(VERSION_0_8)]
  )

  assert manager._server_to_client_schema["defs"] == "server_defs"
  # Basic catalog might have a URI-based ID if not explicitly matched
  # So we check if any catalog exists
  assert len(manager._supported_catalogs) >= 1
  # The first one should be the basic one
  catalog = manager._supported_catalogs[0]
  assert catalog.catalog_schema["version"] == VERSION_0_8
  assert "Text" in catalog.catalog_schema["components"]


def test_schema_manager_init_invalid_version():
  with pytest.raises(ValueError, match="Unknown A2UI specification version"):
    A2uiSchemaManager("invalid_version")


def test_schema_manager_fallback_local_assets(mock_importlib_resources):
  # Force importlib to fail
  # Note: A2UI_ASSET_PACKAGE is "a2ui.assets"
  mock_importlib_resources.side_effect = FileNotFoundError("Package not found")

  with (
      patch("os.path.exists", return_value=True),
      patch("builtins.open", new_callable=MagicMock) as mock_open,
  ):

    def open_side_effect(path, *args, **kwargs):
      path_str = str(path)
      if "server_to_client" in path_str:
        return io.StringIO(
            '{"$schema": "https://json-schema.org/draft/2020-12/schema", "defs":'
            ' "local_server"}'
        )
      elif "standard_catalog" in path_str or "catalog" in path_str:
        return io.StringIO(
            '{"$schema": "https://json-schema.org/draft/2020-12/schema",'
            ' "catalogId": "basic", "components": {"LocalText": {}}}'
        )
      raise FileNotFoundError(path)

    mock_open.side_effect = open_side_effect

    manager = A2uiSchemaManager(
        VERSION_0_8, catalogs=[BasicCatalog.get_config(VERSION_0_8)]
    )

    assert manager._server_to_client_schema["defs"] == "local_server"
    assert len(manager._supported_catalogs) >= 1
    catalog = manager._supported_catalogs[0]
    assert "LocalText" in catalog.catalog_schema["components"]



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants for A2UI parsing."""

DEFAULT_ROOT_ID = "root"

# Message types (v0.8)
MSG_TYPE_BEGIN_RENDERING = "beginRendering"
MSG_TYPE_SURFACE_UPDATE = "surfaceUpdate"
MSG_TYPE_DATA_MODEL_UPDATE = "dataModelUpdate"
MSG_TYPE_DELETE_SURFACE = "deleteSurface"

# Message types (v0.9)
MSG_TYPE_CREATE_SURFACE = "createSurface"
MSG_TYPE_UPDATE_COMPONENTS = "updateComponents"
MSG_TYPE_UPDATE_DATA_MODEL = "updateDataModel"
# deleteSurface is shared between v0.8 and v0.9

# Conversational text (non-A2UI)
MSG_TYPE_TEXT = "text"



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Optional, List, AsyncIterable, TYPE_CHECKING

if TYPE_CHECKING:
  from a2ui.parser.streaming import A2uiStreamParser
from a2a.types import (
    Part,
    DataPart,
    TextPart,
)

logger = logging.getLogger(__name__)

MIME_TYPE_KEY = "mimeType"
A2UI_MIME_TYPE = "application/a2ui+json"
DEPRECATED_A2UI_MIME_TYPE = "application/json+a2ui"


def create_a2ui_part(a2ui_data: dict[str, Any], version: Optional[str] = None) -> Part:
  """Creates an A2A Part containing A2UI data.

  Args:
      a2ui_data: The A2UI data dictionary.
      version: Optional version string.

  Returns:
      An A2A Part with a DataPart containing the A2UI data.
  """
  mime_type = A2UI_MIME_TYPE
  if version in ("0.8", "0.9", "v0.8", "v0.9"):
    mime_type = DEPRECATED_A2UI_MIME_TYPE

  return Part(
      root=DataPart(
          data=a2ui_data,
          metadata={
              MIME_TYPE_KEY: mime_type,
          },
      )
  )


def is_a2ui_part(part: Part) -> bool:
  """Checks if an A2A Part contains A2UI data.

  Args:
      part: The A2A Part to check.

  Returns:
      True if the part contains A2UI data, False otherwise.
  """
  return (
      isinstance(part.root, DataPart)
      and part.root.metadata
      and part.root.metadata.get(MIME_TYPE_KEY)
      in (A2UI_MIME_TYPE, DEPRECATED_A2UI_MIME_TYPE)
  )


def get_a2ui_datapart(part: Part) -> Optional[DataPart]:
  """Extracts the DataPart containing A2UI data from an A2A Part, if present.

  Args:
      part: The A2A Part to extract A2UI data from.

  Returns:
      The DataPart containing A2UI data if present, None otherwise.
  """
  if is_a2ui_part(part):
    return part.root
  return None


def parse_response_to_parts(
    content: str,
    validator: Optional[Any] = None,
    fallback_text: Optional[str] = None,
    version: Optional[str] = None,
) -> List[Part]:
  """Helper to parse LLM response content into A2A Parts, with optional validation.

  Args:
      content: The LLM response content, potentially containing A2UI delimiters.
      validator: Optional validator to run against extracted JSON payloads.
      fallback_text: Optional text to return if no parts are successfully created.
      version: Optional version string.

  Returns:
      A list of A2A Part objects (TextPart and/or DataPart).
  """
  from a2ui.parser.parser import parse_response

  parts = []
  try:
    response_parts = parse_response(content)

    for part in response_parts:
      if part.text:
        parts.append(Part(root=TextPart(text=part.text)))

      if part.a2ui_json:
        json_data = part.a2ui_json
        if validator:
          validator.validate(json_data)

        if isinstance(json_data, list):
          for message in json_data:
            parts.append(create_a2ui_part(message, version=version))
        else:
          parts.append(create_a2ui_part(json_data, version=version))

  except Exception as e:
    logger.warning(f"Failed to parse or validate A2UI response: {e}")

  if not parts and fallback_text:
    parts.append(Part(root=TextPart(text=fallback_text)))

  return parts


async def stream_response_to_parts(
    parser: "A2uiStreamParser",
    token_stream: AsyncIterable[str],
    version: Optional[str] = None,
) -> AsyncIterable[Part]:
  """Helper to parse a stream of LLM tokens into A2A Parts incrementally.

  Args:
      parser: A2uiStreamParser instance to process the stream.
      token_stream: An async iterable of strings (tokens).
      version: Optional version string.

  Yields:
      A2A Part objects as they are discovered in the stream.
  """
  async for token in token_stream:
    logger.info("-----------------------------")
    logger.info(f"--- AGENT: Received token:\n{token}")
    response_parts = parser.process_chunk(token)
    logger.info(
        f"--- AGENT: Response parts:\n{[part.a2ui_json for part in response_parts]}\n"
    )
    logger.info("-----------------------------")

    for part in response_parts:
      if part.text:
        yield Part(root=TextPart(text=part.text))

      if part.a2ui_json:
        json_data = part.a2ui_json

        if isinstance(json_data, list):
          for message in json_data:
            yield create_a2ui_part(message, version=version)
        else:
          yield create_a2ui_part(json_data, version=version)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for A2UI Schema manipulation."""

import json
import logging
import os
import importlib.resources
from typing import Any, Dict

from .constants import A2UI_ASSET_PACKAGE, SPECIFICATION_DIR, ENCODING
from .catalog_provider import FileSystemCatalogProvider


def find_repo_root(start_path: str) -> str | None:
  """Finds the repository root by looking for the 'specification' directory."""
  current = os.path.abspath(start_path)
  while True:
    if os.path.isdir(os.path.join(current, SPECIFICATION_DIR)):
      return current
    parent = os.path.dirname(current)
    if parent == current:
      return None
    current = parent


def load_from_bundled_resource(
    version: str,
    resource_key: str,
    spec_map: Dict[str, Dict[str, str]],
) -> Dict[str, Any]:
  """Loads a schema resource from bundled package resources."""
  spec_map = spec_map.get(version)
  if not spec_map:
    raise ValueError(f"Unknown A2UI version: {version}")

  if resource_key not in spec_map:
    return None

  rel_path = spec_map[resource_key]
  filename = os.path.basename(rel_path)

  # 1. Try to load from the bundled package resources.
  try:
    traversable = importlib.resources.files(A2UI_ASSET_PACKAGE)
    traversable = traversable.joinpath(version).joinpath(filename)
    with traversable.open("r", encoding=ENCODING) as f:
      return json.load(f)
  except Exception as e:
    logging.debug("Could not load '%s' from package resources: %s", filename, e)

  # 2. Fallback to local assets
  # This handles cases where assets might be present in src but not installed
  try:
    # The assets are located at a2ui/assets/<version>/<filename>
    # This file is at a2ui/inference/schema/manager.py
    # So, we need to go up 3 directories to 'a2ui', then down to 'assets'
    potential_path = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "assets",
            version,
            filename,
        )
    )
    if os.path.exists(potential_path):
      provider = FileSystemCatalogProvider(potential_path)
      return provider.load()
  except Exception as e:
    logging.debug("Could not load schema '%s' from local assets: %s", filename, e)

  # 3. Fallback: Source Repository (specification/...)
  # This handles cases where we are running directly from source tree
  # And assets are not yet copied to src/a2ui/assets
  # manager.py is at a2a_agents/python/a2ui_agent/src/a2ui/inference/schema/manager.py
  # Dynamically find repo root by looking for "specification" directory
  try:
    repo_root = find_repo_root(os.path.dirname(__file__))
    if repo_root:
      source_path = os.path.join(repo_root, rel_path)
      if os.path.exists(source_path):
        provider = FileSystemCatalogProvider(source_path)
        return provider.load()
  except Exception as e:
    logging.debug("Could not load schema from source repo: %s", e)

  raise IOError(f"Could not load schema {filename} for version {version}")


# LLM is instructed to generate a list of messages, so we wrap the bundled schema in an array.
def wrap_as_json_array(a2ui_schema: dict[str, Any]) -> dict[str, Any]:
  """Wraps the A2UI schema in an array object to support multiple parts.

  Args:
      a2ui_schema: The A2UI schema to wrap.

  Returns:
      The wrapped A2UI schema object.

  Raises:
      ValueError: If the A2UI schema is empty.
  """
  if not a2ui_schema:
    raise ValueError("A2UI schema is empty")
  return {"type": "array", "items": a2ui_schema}


def deep_update(d: dict, u: dict) -> dict:
  """Recursively update a dict with another dict."""
  for k, v in u.items():
    if isinstance(v, dict):
      d[k] = deep_update(d.get(k, {}), v)
    else:
      d[k] = v
  return d



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the A2uiEventConverter class."""

from unittest.mock import MagicMock, patch

import pytest

from a2ui.adk.a2a.event_converter import A2uiEventConverter
from a2ui.adk.a2a.part_converter import A2uiPartConverter
from a2ui.schema.catalog import A2uiCatalog


def test_event_converter_injects_catalog():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  event_mock = MagicMock()
  invocation_context_mock = MagicMock()
  # Correctly access session via mock
  invocation_context_mock.session.state = {"system:a2ui_catalog": catalog_mock}

  converter = A2uiEventConverter()

  with patch(
      "google.adk.a2a.converters.event_converter.convert_event_to_a2a_events"
  ) as mock_base_converter:
    mock_base_converter.return_value = []

    # Converter is not async
    converter(event_mock, invocation_context_mock)

    # Verify that mock_base_converter was called with a part_converter that uses the catalog
    args, kwargs = mock_base_converter.call_args
    effective_part_converter = args[4]

    assert effective_part_converter.__name__ == "convert"
    assert isinstance(effective_part_converter.__self__, A2uiPartConverter)
    assert effective_part_converter.__self__._catalog == catalog_mock


def test_event_converter_falls_back_without_catalog():
  event_mock = MagicMock()
  invocation_context_mock = MagicMock()
  invocation_context_mock.session.state = {}  # No catalog

  converter = A2uiEventConverter()

  with patch(
      "google.adk.a2a.converters.event_converter.convert_event_to_a2a_events"
  ) as mock_base_converter:
    mock_base_converter.return_value = []

    # Converter is not async
    converter(event_mock, invocation_context_mock)

    args, kwargs = mock_base_converter.call_args
    effective_part_converter = args[4]

    from google.adk.a2a.converters.part_converter import convert_genai_part_to_a2a_part

    assert effective_part_converter == convert_genai_part_to_a2a_part


def test_event_converter_propagates_fallback_text():
  catalog_mock = MagicMock(spec=A2uiCatalog)
  event_mock = MagicMock()
  invocation_context_mock = MagicMock()
  invocation_context_mock.session.state = {"system:a2ui_catalog": catalog_mock}

  custom_fallback = "Custom event fallback text"
  converter = A2uiEventConverter(fallback_text=custom_fallback)

  with patch(
      "google.adk.a2a.converters.event_converter.convert_event_to_a2a_events"
  ) as mock_base_converter:
    mock_base_converter.return_value = []

    converter(event_mock, invocation_context_mock)

    args, kwargs = mock_base_converter.call_args
    effective_part_converter = args[4]

    assert isinstance(effective_part_converter.__self__, A2uiPartConverter)
    assert effective_part_converter.__self__._fallback_text == custom_fallback



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for providing A2UI catalog schemas and resources."""

import json
from abc import ABC, abstractmethod
from json.decoder import JSONDecodeError
from typing import Any, Dict
from .constants import ENCODING


class A2uiCatalogProvider(ABC):
  """Abstract base class for providing A2UI schemas and catalogs."""

  @abstractmethod
  def load(self) -> Dict[str, Any]:
    """Loads a catalog definition.

    Returns:
      The loaded catalog as a dictionary.
    """
    pass


class FileSystemCatalogProvider(A2uiCatalogProvider):
  """Loads catalog definition from the local filesystem."""

  def __init__(self, path: str):
    self.path = path

  def load(self) -> Dict[str, Any]:
    try:
      with open(self.path, "r", encoding=ENCODING) as f:
        return json.load(f)
    except (FileNotFoundError, JSONDecodeError) as e:
      raise IOError(f"Could not load schema from {self.path}: {e}") from e



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy
import glob
import json
import logging
import os
from dataclasses import dataclass, field, replace
from typing import Any, Dict, List, Optional, TYPE_CHECKING
from urllib.parse import urlparse

from .catalog_provider import A2uiCatalogProvider, FileSystemCatalogProvider
from .constants import (
    A2UI_SCHEMA_BLOCK_START,
    A2UI_SCHEMA_BLOCK_END,
    CATALOG_COMPONENTS_KEY,
    CATALOG_ID_KEY,
    VERSION_0_8,
    ENCODING,
)


@dataclass
class CatalogConfig:
  """
  Configuration for a catalog of components.

  A catalog consists of a provider that knows how to load the schema,
  and optionally a path or glob pattern to examples.

  Attributes:
    name: The name of the catalog.
    provider: The provider to use to load the catalog schema.
    examples_path: The path or glob pattern to the examples.
  """

  name: str
  provider: A2uiCatalogProvider
  examples_path: Optional[str] = None

  @classmethod
  def from_path(
      cls, name: str, catalog_path: str, examples_path: Optional[str] = None
  ) -> "CatalogConfig":
    """Returns a CatalogConfig that loads from a local path or 'file://' URI."""
    parsed = urlparse(catalog_path)
    if not parsed.scheme or parsed.scheme == "file":
      catalog_provider = FileSystemCatalogProvider(parsed.path)
    elif parsed.scheme in ["http", "https"]:
      raise NotImplementedError("HTTP support is coming soon.")
    else:
      raise ValueError(f"Unsupported catalog URL scheme: {catalog_path}")

    return cls(
        name=name,
        provider=catalog_provider,
        examples_path=resolve_examples_path(examples_path),
    )


def resolve_examples_path(path: Optional[str]) -> Optional[str]:
  if path:
    parsed = urlparse(path)
    if not parsed.scheme or parsed.scheme == "file":
      return parsed.path
    else:
      raise ValueError(f"Unsupported examples URL scheme: {path}")
  return None


def _collect_refs(obj: Any) -> set[str]:
  """Recursively collects all $ref values from a JSON object."""
  refs = set()
  if isinstance(obj, dict):
    for k, v in obj.items():
      if k == "$ref" and isinstance(v, str):
        refs.add(v)
      else:
        refs.update(_collect_refs(v))
  elif isinstance(obj, list):
    for item in obj:
      refs.update(_collect_refs(item))
  return refs


def _prune_defs_by_reachability(
    defs: Dict[str, Any],
    root_def_names: List[str],
    internal_ref_prefix: str = "#/$defs/",
) -> Dict[str, Any]:
  """Prunes definitions not reachable from the provided roots.

  Args:
    defs: The dictionary of definitions to prune.
    root_def_names: The names of the definitions to start the traversal from.
    internal_ref_prefix: The prefix used for internal references.

  Returns:
    A new dictionary containing only reachable definitions.
  """
  visited_defs = set()
  refs_queue = collections.deque(root_def_names)

  while refs_queue:
    def_name = refs_queue.popleft()
    if def_name in defs and def_name not in visited_defs:
      visited_defs.add(def_name)

      internal_refs = _collect_refs(defs[def_name])
      for ref in internal_refs:
        if ref.startswith(internal_ref_prefix):
          refs_queue.append(ref.split(internal_ref_prefix)[-1])

  return {k: v for k, v in defs.items() if k in visited_defs}


@dataclass(frozen=True)
class A2uiCatalog:
  """Represents a processed component catalog with its schema.

  Attributes:
    version: The version of the catalog.
    name: The name of the catalog.
    s2c_schema: The server-to-client schema.
    common_types_schema: The common types schema.
    catalog_schema: The catalog schema.
  """

  version: str
  name: str
  s2c_schema: Dict[str, Any]
  common_types_schema: Dict[str, Any]
  catalog_schema: Dict[str, Any]

  @property
  def catalog_id(self) -> str:
    if CATALOG_ID_KEY not in self.catalog_schema:
      raise ValueError(f"Catalog '{self.name}' missing catalogId")
    return self.catalog_schema[CATALOG_ID_KEY]

  @property
  def validator(self) -> "A2uiValidator":
    from .validator import A2uiValidator

    return A2uiValidator(self)

  def _with_pruned_components(self, allowed_components: List[str]) -> "A2uiCatalog":
    """Returns a new catalog with only allowed components.

    Args:
      allowed_components: List of component names to include.

    Returns:
      A copy of the catalog with only allowed components.
    """

    if not allowed_components:
      return self

    schema_copy = copy.deepcopy(self.catalog_schema)

    if CATALOG_COMPONENTS_KEY in schema_copy and isinstance(
        schema_copy[CATALOG_COMPONENTS_KEY], dict
    ):
      all_comps = schema_copy[CATALOG_COMPONENTS_KEY]
      schema_copy[CATALOG_COMPONENTS_KEY] = {
          k: v for k, v in all_comps.items() if k in allowed_components
      }

    # Filter anyComponent oneOf if it exists
    # Path: $defs -> anyComponent -> oneOf
    if "$defs" in schema_copy and "anyComponent" in schema_copy["$defs"]:
      any_comp = schema_copy["$defs"]["anyComponent"]
      if "oneOf" in any_comp and isinstance(any_comp["oneOf"], list):
        filtered_one_of = []
        for item in any_comp["oneOf"]:
          if "$ref" in item:
            ref = item["$ref"]
            if ref.startswith(f"#/{CATALOG_COMPONENTS_KEY}/"):
              comp_name = ref.split("/")[-1]
              if comp_name in allowed_components:
                filtered_one_of.append(item)
            else:
              logging.warning(f"Skipping unknown ref format: {ref}")
          else:
            logging.warning(f"Skipping non-ref item in anyComponent oneOf: {item}")

        any_comp["oneOf"] = filtered_one_of

    return replace(self, catalog_schema=schema_copy)

  def _with_pruned_messages(self, allowed_messages: List[str]) -> "A2uiCatalog":
    """Returns a new catalog with only allowed messages.

    Args:
      allowed_messages: List of message names to include in s2c_schema.

    Returns:
      A copy of the catalog with only allowed messages.
    """
    if not allowed_messages:
      return self

    s2c_schema_copy = copy.deepcopy(self.s2c_schema)

    if self.version == VERSION_0_8:
      # 0.8 style: Messages are in root properties.
      if "properties" in s2c_schema_copy and isinstance(
          s2c_schema_copy["properties"], dict
      ):
        s2c_schema_copy["properties"] = _prune_defs_by_reachability(
            defs=s2c_schema_copy["properties"],
            root_def_names=allowed_messages,
            internal_ref_prefix="#/properties/",
        )
    else:
      # 0.9+ style: Messages are in $defs and referenced via oneOf.
      if "oneOf" in s2c_schema_copy and isinstance(s2c_schema_copy["oneOf"], list):
        s2c_schema_copy["oneOf"] = [
            item
            for item in s2c_schema_copy["oneOf"]
            if "$ref" in item
            and item["$ref"].startswith("#/$defs/")
            and item["$ref"].split("/")[-1] in allowed_messages
        ]

      if "$defs" in s2c_schema_copy and isinstance(s2c_schema_copy["$defs"], dict):
        s2c_schema_copy["$defs"] = _prune_defs_by_reachability(
            defs=s2c_schema_copy["$defs"],
            root_def_names=allowed_messages,
            internal_ref_prefix="#/$defs/",
        )

    return replace(self, s2c_schema=s2c_schema_copy)

  def with_pruning(
      self,
      allowed_components: Optional[List[str]] = None,
      allowed_messages: Optional[List[str]] = None,
  ) -> "A2uiCatalog":
    """Returns a new catalog with pruned components and messages.

    Args:
      allowed_components: List of component names to include.
      allowed_messages: List of message names to include in s2c_schema.

    Returns:
      A copy of the catalog with pruned components and messages.
    """
    catalog = self
    if allowed_components:
      catalog = catalog._with_pruned_components(allowed_components)

    if allowed_messages:
      catalog = catalog._with_pruned_messages(allowed_messages)

    return catalog._with_pruned_common_types()

  def _with_pruned_common_types(self) -> "A2uiCatalog":
    """Returns a new catalog with unused common types pruned from the schema."""
    if not self.common_types_schema or "$defs" not in self.common_types_schema:
      return self

    # Initialize roots with ONLY refs targeting common_types.json from external schemas
    external_refs = _collect_refs(self.catalog_schema)
    external_refs.update(_collect_refs(self.s2c_schema))

    root_common_types = []
    for ref in external_refs:
      if "common_types.json#/$defs/" in ref:
        root_common_types.append(ref.split("#/$defs/")[-1])

    new_common_types_schema = copy.deepcopy(self.common_types_schema)
    new_common_types_schema["$defs"] = _prune_defs_by_reachability(
        defs=new_common_types_schema["$defs"],
        root_def_names=root_common_types,
    )

    return replace(self, common_types_schema=new_common_types_schema)

  def render_as_llm_instructions(self) -> str:
    """Renders the catalog and schema as LLM instructions."""
    all_schemas = []
    all_schemas.append(A2UI_SCHEMA_BLOCK_START)

    server_client_str = (
        json.dumps(self.s2c_schema, separators=(",", ":")) if self.s2c_schema else "{}"
    )
    all_schemas.append(f"### Server To Client Schema:\n{server_client_str}")

    if (
        self.common_types_schema
        and "$defs" in self.common_types_schema
        and self.common_types_schema["$defs"]
    ):
      common_str = json.dumps(self.common_types_schema, separators=(",", ":"))
      all_schemas.append(f"### Common Types Schema:\n{common_str}")

    catalog_str = json.dumps(self.catalog_schema, separators=(",", ":"))
    all_schemas.append(f"### Catalog Schema:\n{catalog_str}")

    all_schemas.append(A2UI_SCHEMA_BLOCK_END)

    return "\n\n".join(all_schemas)

  def load_examples(self, path: Optional[str], validate: bool = False) -> str:
    """Loads and validates examples from a directory or a glob pattern."""
    if not path:
      return ""

    # If it's a directory, support backward compatibility by appending /*.json
    if os.path.isdir(path):
      pattern = os.path.join(path, "*.json")
    else:
      pattern = path

    # Use glob to find files
    matched_files = glob.glob(pattern, recursive=True)

    if not matched_files:
      if not os.path.isdir(path) and not any(c in path for c in "*?[]"):
        logging.warning(
            f"Example path {path} is neither a directory nor a valid glob pattern"
        )
      return ""

    # Sort for determinism
    matched_files.sort()

    merged_examples = []
    for full_path in matched_files:
      if not os.path.isfile(full_path):
        continue
      basename = os.path.splitext(os.path.basename(full_path))[0]
      with open(full_path, "r", encoding=ENCODING) as f:
        content = f.read()

      if validate:
        self._validate_example(full_path, content)

      merged_examples.append(f"---BEGIN {basename}---\n{content}\n---END {basename}---")

    if not merged_examples:
      return ""
    return "\n\n".join(merged_examples)

  def _validate_example(self, full_path: str, content: str) -> None:
    try:
      json_data = json.loads(content)
      self.validator.validate(json_data)
    except Exception as e:
      raise ValueError(f"Failed to validate example {full_path}: {e}") from e



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import json
from typing import Any, List, Dict, Optional, Set

from .streaming import A2uiStreamParser
from .response_part import ResponsePart
from .constants import *
from ..schema.constants import VERSION_0_8, SURFACE_ID_KEY, CATALOG_COMPONENTS_KEY


class A2uiStreamParserV08(A2uiStreamParser):
  """Streaming parser implementation for A2UI v0.8 specification."""

  def __init__(self, catalog=None):
    super().__init__(catalog=catalog)
    self._yielded_begin_rendering_surfaces: Set[str] = set()

  @property
  def _placeholder_component(self) -> Dict[str, Any]:
    """Returns the placeholder component."""
    return {
        'component': {
            'Row': {
                'children': {'explicitList': []},
            }
        }
    }

  @property
  def _yielded_surfaces_set(self) -> Set[str]:
    """Provides access to version-specific yielded surfaces set."""
    return self._yielded_begin_rendering_surfaces

  def is_protocol_msg(self, obj: Dict[str, Any]) -> bool:
    """Checks if the object is a recognized v0.8 message."""
    return any(
        k in obj
        for k in (
            MSG_TYPE_BEGIN_RENDERING,
            MSG_TYPE_SURFACE_UPDATE,
            MSG_TYPE_DATA_MODEL_UPDATE,
            MSG_TYPE_DELETE_SURFACE,
        )
    )

  @property
  def _data_model_msg_type(self) -> str:
    """Returns the message type identifier for data model updates."""
    return MSG_TYPE_DATA_MODEL_UPDATE

  def _sniff_metadata(self):
    """Sniffs for v0.8 metadata in the json_buffer."""

    def get_latest_value(key: str) -> Optional[str]:
      idx = len(self._json_buffer)
      while True:
        idx = self._json_buffer.rfind(f'"{key}"', 0, idx)
        if idx == -1:
          return None
        match = re.match(rf'"{key}"\s*:\s*"([^"]+)"', self._json_buffer[idx:])
        if match:
          return match.group(1)

    self.surface_id = get_latest_value('surfaceId')

    parsed_root = get_latest_value('root')
    if parsed_root is not None:
      self.root_id = parsed_root

    if f'"{MSG_TYPE_BEGIN_RENDERING}":' in self._json_buffer:
      self.add_msg_type(MSG_TYPE_BEGIN_RENDERING)
    if f'"{MSG_TYPE_SURFACE_UPDATE}":' in self._json_buffer:
      self.add_msg_type(MSG_TYPE_SURFACE_UPDATE)
    if f'"{MSG_TYPE_DATA_MODEL_UPDATE}":' in self._json_buffer:
      self.add_msg_type(MSG_TYPE_DATA_MODEL_UPDATE)
    if f'"{MSG_TYPE_DELETE_SURFACE}":' in self._json_buffer:
      self.add_msg_type(MSG_TYPE_DELETE_SURFACE)

  def _handle_complete_object(
      self,
      obj: Dict[str, Any],
      sid: Optional[str],
      messages: List[ResponsePart],
  ) -> bool:
    """Handles v0.8 specific complete objects."""
    if not isinstance(obj, dict):
      return False

    if self._validator:
      self._validator.validate(obj, root_id=sid, strict_integrity=False)

    # Update state based on the message content
    surface_id = obj.get(SURFACE_ID_KEY, self.surface_id)
    if MSG_TYPE_SURFACE_UPDATE in obj:
      val = obj[MSG_TYPE_SURFACE_UPDATE]
      if isinstance(val, dict):
        surface_id = val.get(SURFACE_ID_KEY) or surface_id
    elif MSG_TYPE_BEGIN_RENDERING in obj:
      val = obj[MSG_TYPE_BEGIN_RENDERING]
      if isinstance(val, dict):
        surface_id = val.get(SURFACE_ID_KEY) or surface_id
    elif MSG_TYPE_DELETE_SURFACE in obj:
      val = obj[MSG_TYPE_DELETE_SURFACE]
      if isinstance(val, str):
        surface_id = val
      elif isinstance(val, dict):
        surface_id = val.get(SURFACE_ID_KEY) or surface_id

    self.surface_id = surface_id
    sid = self.surface_id or 'unknown'

    if MSG_TYPE_DELETE_SURFACE in obj:
      if sid in self._yielded_surfaces_set or self._buffered_start_message:
        self._delete_surface(sid)

    if sid in self._deleted_surfaces:
      return True

    if (
        (MSG_TYPE_SURFACE_UPDATE in obj or MSG_TYPE_DELETE_SURFACE in obj)
        and sid not in self._yielded_surfaces_set
        and not self._buffered_start_message
    ):
      if sid not in self._pending_messages:
        self._pending_messages[sid] = []
      self._pending_messages[sid].append(obj)
      return True

    if MSG_TYPE_BEGIN_RENDERING in obj:
      br_val = obj[MSG_TYPE_BEGIN_RENDERING]
      if isinstance(br_val, dict):
        self.surface_id = br_val.get(SURFACE_ID_KEY, self.surface_id)
      self.root_id = br_val.get('root', self.root_id or DEFAULT_ROOT_ID)
      self._buffered_start_message = obj

      # Yield beginRendering immediately when it completes
      if sid not in self._yielded_start_messages:
        self._yield_messages([obj], messages)
        self._yielded_start_messages.add(sid)
        self._yielded_surfaces_set.add(sid)
        self._buffered_start_message = None

      if sid in self._pending_messages:
        pending_list = self._pending_messages.pop(sid)
        for pending_msg in pending_list:
          self._handle_complete_object(pending_msg, sid, messages)

      self.yield_reachable(messages)
      return True

    if MSG_TYPE_SURFACE_UPDATE in obj:
      self.add_msg_type(MSG_TYPE_SURFACE_UPDATE)
      components = obj[MSG_TYPE_SURFACE_UPDATE].get('components', [])
      for comp in components:
        if isinstance(comp, dict) and 'id' in comp:
          self._seen_components[comp['id']] = comp
      self.yield_reachable(messages, check_root=True, raise_on_orphans=False)
      return True

    if MSG_TYPE_DATA_MODEL_UPDATE in obj:
      self.add_msg_type(MSG_TYPE_DATA_MODEL_UPDATE)
      self.update_data_model(obj[MSG_TYPE_DATA_MODEL_UPDATE], messages)
      self._yield_messages([obj], messages)
      self.yield_reachable(messages, check_root=False, raise_on_orphans=False)
      return True

    if MSG_TYPE_DELETE_SURFACE in obj:
      self._yield_messages([obj], messages)
      return True

    # If unknown, let base class yield it or yield it here
    self._yield_messages([obj], messages)
    return True

  def _construct_partial_message(
      self, processed_components: List[Dict[str, Any]], active_msg_type: str
  ) -> Dict[str, Any]:
    """Constructs a partial message for v0.8 (always surfaceUpdate)."""
    payload = {
        SURFACE_ID_KEY: self.surface_id,
        CATALOG_COMPONENTS_KEY: processed_components,
    }
    return {MSG_TYPE_SURFACE_UPDATE: payload}

  def _get_active_msg_type_for_components(self) -> Optional[str]:
    """Determines which msg_type to use when wrapping component updates."""
    if self._active_msg_type:
      return self._active_msg_type
    for mt in self._msg_types:
      if mt in (MSG_TYPE_SURFACE_UPDATE, MSG_TYPE_BEGIN_RENDERING):
        self._active_msg_type = mt
        return mt
    return self._msg_types[0] if self._msg_types else None

  def _deduplicate_data_model(self, m: Dict[str, Any], strict_integrity: bool) -> bool:
    if MSG_TYPE_DATA_MODEL_UPDATE in m:
      dm = m[MSG_TYPE_DATA_MODEL_UPDATE]
      raw_contents = dm.get('contents', {})
      contents_dict = {}
      if isinstance(raw_contents, list):
        for entry in raw_contents:
          if isinstance(entry, dict) and 'key' in entry:
            key = entry['key']
            val = (
                entry.get('valueString')
                or entry.get('valueNumber')
                or entry.get('valueBoolean')
                or entry.get('valueMap')
            )
            if key and val is not None:
              contents_dict[key] = val
      elif isinstance(raw_contents, dict):
        contents_dict = raw_contents

      if contents_dict:
        is_new = False
        for k, v in contents_dict.items():
          if self._yielded_data_model.get(k) != v:
            is_new = True
            break
        if not is_new and strict_integrity:
          return False
        self._yielded_data_model.update(contents_dict)
    return True
